//---------------------------------------------------------------------------
//
//  Module: vddgrab.c
//
//  Purpose:
//     VDDVGA grabber support
//
//     Based on original Windows 3.1 VDD sources as supplied by
//     the Microsoft Windows 3.1 Device Driver Development Kit.
//
//  Development Team:
//     Bryan A. Woodruff
//
//  History:   Date       Author      Comment
//              4/25/93   BryanW      Rewritten in 'C'.
//
//---------------------------------------------------------------------------
//
//               Copyright (c) 1993 Woodruff Software Systems.
//                          All Rights Reserved.
//
//---------------------------------------------------------------------------

#include <vmm.h>
#include "vddvga.h"
#include "vddphmem.h"
#include "vddvmmem.h"

#pragma data_seg( "_LDATA", "CODE" ) 

//=============================================================================
//                    G L O B A L   D A T A   A R E A 
//=============================================================================

//=============================================================================
//                             L C O D E
//=============================================================================

//--------------------------------------------------------------------------
//  
//  BOOL VDD_Grab_DoGrab
//  
//  Description:
//      Allocate copy memory, call VDD_Grab_Save_Dirty_Pages
//      to force save memory to contain all resent changes, then
//      copy current video memory state from save memory to copy
//      memory.
//  
//  Parameters:
//      DWORD hVM
//         handle to VM
//  
//      PVDDCB pVMCB
//         pointer to VM control block
//  
//  Return (BOOL):
//      FALSE if mem alloc failed or other failure.
//  
//  History:   Date       Author      Comment
//             10/27/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

BOOL VDD_Grab_DoGrab
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{
   if (hVM != hVidFocusVM)
   {
      DEBUGMSG( "GRAB for VM other than focus - should not get this - GRAB IGNORED!!\r\n" ) ;
      return FALSE ;
   }

   if (!VDD_Grab_ACopy( hVM, pVMCB ))
   {
      DEBUGMSG( "VDD: failed to allocate memory for grab!\r\n" ) ;
      VDD_Grab_DCopy( hVM, pVMCB ) ;
      return FALSE ;
   }

   VDD_State_ScrnOff( hVM ) ;

   if (pVMCB -> VDD_Flags & fVDD_NoSaveRes)
   {
      UINT   i, uVisFirst, uVisPages ;
      PBYTE  pDst ;

      // auto res disabled - abnormal grab

      // abnormal grab is done when a VM has done the INT2F
      // to inform us that it CanRestore and AutoRestore has
      // been disabled with the INI switch.  In this case the
      // VM does not have any save memory allocated, so we copy
      // MSttCopy into MStt so that the copy memory looks like
      // save memory.  Then we call VDD_VM_Mem_Save_Page for all
      // visible pages to copy the physical video memory state
      // into the copy memory, and then we clear MStt.

      pVMCB -> VDD_MStt = pVMCB -> VDD_MSttCopy ;

      VDD_PHMem_GetVisiblePages( hVM, pVMCB, &uVisFirst, &uVisPages ) ;

      // call save page for all visible pages to save video
      // page memory into copy memory.

      for (i = 0; i < uVisPages; i++)
      {
         if (pVMCB -> VDD_PageMap[ uVisFirst + i ])
            VDD_VMMem_SavePage( hVM, pVMCB,
                                pVMCB -> VDD_PageMap[ uVisFirst + i ] ) ;

      }

      // clear VDD_MSttCopy

      pDst = (PBYTE) &(pVMCB -> VDD_MStt) ;
      for (i = 0; i < sizeof( MEMSTATE); i++)
         *pDst = 0 ;
   }
   else
   {
      UINT    i, j ;
      PDWORD  pSrc, pDst ;

      // save any dirty video pages into the VM's video save memory

      VDD_Grab_SaveDirtyPages( hVM, pVMCB ) ;

      // copy save memory to copy memory

      for (i = 0; i < 4; i++)
      {
         if (pVMCB -> VDD_MSttCopy.MSS_planes[ i ].MSS_plane_size)
         {
            pSrc =
               (PDWORD) pVMCB -> VDD_MStt.MSS_planes[ i ].MSS_plane_addr ;

            pDst =
               (PDWORD) pVMCB -> VDD_MSttCopy.MSS_planes[ i ].MSS_plane_addr ;

            for (j = 0;
                 j < pVMCB -> VDD_MSttCopy.MSS_planes[ i ].MSS_plane_size << 10;
                 j++)
               *pDst++ = *pSrc++ ;
         }
      }
   }

   VDD_State_ScrnOn( hVM ) ;
   return TRUE ;


} // VDD_Grab_DoGrab()

//--------------------------------------------------------------------------
//  
//  BOOL VDD_Grab_ACopy
//  
//  Description:
//      Allocate copy memory
//  
//  Parameters:
//      DWORD hVM
//         handle to VM
//  
//      PVDDCB pVMCB
//         pointer to VM control block
//  
//  Return (BOOL):
//      TRUE on success
//  
//  History:   Date       Author      Comment
//             10/27/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

BOOL VDD_Grab_ACopy
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{
   VDD_VMMem_AdjustPlaneSizes( hVM, pVMCB, &pVMCB -> VDD_MSttCopy, TRUE ) ;
   return (VDD_VMMem_ReAlloc( hVM, pVMCB, &pVMCB -> VDD_MSttCopy )) ;

} // VDD_Grab_ACopy()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Grab_DCopy
//  
//  Description:
//      Deallocate copy memory.
//  
//  Parameters:
//      DWORD hVM
//         handle to VM
//  
//      PVDDCB pVMCB
//         pointer to VM control block
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/27/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_Grab_DCopy
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{
   UNUSED_PARAM( hVM ) ;

   VDD_VMMem_DeAlloc( &pVMCB -> VDD_MSttCopy ) ;

} // VDD_Grab_DCopy()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Grab_GetTextCopyBuf
//  
//  Description:
//      Allocate a copy memory buffer for the grabber.
//  
//  Parameters:
//      DWORD hVM
//         handle to VM 
//  
//      PVDDCB pVMCB
//         pointer to VM control block
//  
//      PDWORD pAddr
//         pointer to address
//  
//      PDWORD pSize
//         pointer to size
//  
//  Return (VOID):
//      *pAddr
//         copy buffer address
//
//      *pSize
//         copy buffer size
//  
//  History:   Date       Author      Comment
//             10/27/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_Grab_GetTextCopyBuf
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    PDWORD          pAddr,
    PDWORD          pSize
)
{

   UINT           uVisFirst, uVisPages ;
   PMSSPLANEINFO  pMPI ;

   VDD_PHMem_GetVisiblePages( hVM, pVMCB, &uVisFirst, &uVisPages ) ;

   pMPI = &pVMCB -> VDD_Copy_Buf ;

   if ((uVisPages != pMPI -> MSS_plane_alloced) ||
       (0 == pMPI -> MSS_plane_addr))
   {
      // need to increase plane size or shrink it?

      if ((uVisPages > pMPI -> MSS_plane_alloced) ||
          (0 == (pVMCB -> VDD_PIF & fVidRetainAllo)))
      {
         pMPI -> MSS_plane_size = uVisPages ;
         VDD_VMMem_ReAllocPlane( hVM, pMPI ) ;
      }
   }

   *pAddr = pMPI -> MSS_plane_addr ;
   *pSize = pMPI -> MSS_plane_alloced ;

} // VDD_Grab_GetTextCopyBuf()

//------------------------------------------------------------------------
//  VOID VDD_Grab_DeleteTextCopyBuf
//
//  Description:
//     Free grabber copy memory buffer, if it exists.
//
//  Parameters:
//     DWORD hVM
//        handle to VM
//
//     PVDDCB pVMCB
//        pointer to VM control block
//
//  History:   Date       Author      Comment
//              4/25/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_Grab_DeleteTextCopyBuf
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{
   DWORD hMem ;

   UNUSED_PARAM( hVM ) ;

   TRACEMSG( "Delete copy buffer" ) ;

   if (0 == (pVMCB -> VDD_PIF & fVidRetainAllo))
   {
      pVMCB -> VDD_Copy_Buf.MSS_plane_addr = NULL ;
      pVMCB -> VDD_Copy_Buf.MSS_plane_size = 0 ;
      pVMCB -> VDD_Copy_Buf.MSS_plane_alloced = 0 ;
      hMem = pVMCB -> VDD_Copy_Buf.MSS_plane_handle ;
      pVMCB -> VDD_Copy_Buf.MSS_plane_handle = NULL ;

      if (hMem)
         vmmPageFree( hMem, 0 ) ;
   }

} // end of VDD_Grab_DeleteTextCopyBuf()

//--------------------------------------------------------------------------
//  
//  BOOL VDD_Grab_ChangeQuery
//  
//  Description:
//      Check if any mods recorded in mod array, or if any
//      mapped pages have been dirtied, to determine if
//      the VM's window should be updated.
//  
//  Parameters:
//      DWORD hVM
//         handle to VM
//  
//      PVDDCB pVMCB
//         pointer to VM control block
//  
//  Return (BOOL):
//      TRUE if update needed
//  
//  History:   Date       Author      Comment
//             10/28/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

#define GRAB_MAX_PTES 0x10

BOOL VDD_Grab_ChangeQuery
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{

   DWORD  dwLinPage, aPTEs[ GRAB_MAX_PTES ] ;
   UINT   i, uPageNum, uVisFirst, uVisPages ;

   if (*pVMCB -> VDD_VRAM_Mods)
      return TRUE ;

   if (pVMCB -> VDD_Mod_Rect != 0xFFFF0000)
      return TRUE ;

   VDD_PHMem_GetVisiblePages( hVM, pVMCB, &uVisFirst, &uVisPages ) ;

   if (uVisPages > GRAB_MAX_PTES)
   {
      DEBUGMSG( "VDD: uh oh!! visible pages > GRAB_MAX_PTES\r\n" ) ;
      return TRUE ;
   }
   
   uPageNum = VDD_PHMem_CalcPageAddr( hVM, pVMCB, uVisFirst ) ;
   dwLinPage = ((uPageNum << 12) + ((PVMMCB) hVM) -> CB_High_Linear) >> 12 ;

   vmmCopyPageTable( dwLinPage, uVisPages, aPTEs, 0 ) ;

   // walk the PTEs to see if anything has been touched...

   for (i = 0; i < uVisPages; i++)
      if (aPTEs[ i ] & P_DIRTY)
         return TRUE ;

   return FALSE ;

} // VDD_Grab_ChangeQuery()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Grab_ClearModState
//  
//  Description:
//      Clear the screen modified bit in the VM flags (VDD_Flags)
//      and clear the VDD_VRAM_Mods table.
//  
//  Parameters:
//      DWORD hVM
//         handle to VM
//  
//      PVDDCB pVMCB
//         pointer to VM control block
//  
//  Return (VOID):
//  
//  History:   Date       Author      Comment
//             10/27/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_Grab_ClearModState
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{
   UINT    i ;

   UNUSED_PARAM( hVM ) ;

   for (i = 0; i < VRAM_Mod_List_Size / 4; i++)
      pVMCB -> VDD_Grab_Mods[ i ] = 0 ;

} // VDD_Grab_ClearModState()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Grab_SaveDirtyPages
//  
//  Description:
//      Walk thru list of all video pages owned by a VM, force
//      any dirtied pages to be saved to the VM's save memory
//      and clear all dirty bits.
//
//  Parameters:
//      DWORD hVM
//         handle to VM
//  
//      PVDDCB pVMCB
//         pointer to VM control block
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/28/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_Grab_SaveDirtyPages
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{
   UINT   i ;
   DWORD  hPage ;

   for (i = 0; i < pVMCB -> VDD_VideoPages; i++)
   {
      hPage = pVMCB -> VDD_PageMap[ i ] ;
      if (hPage)
      {
         if (VDD_VMMem_IsPageDirty( hVM, hPage ))
         {
            DWORD dwLinPage ;

            // copy video memory to VM's save memory

            VDD_State_BeginSRSection() ;
            VDD_VMMem_SavePage( hVM, pVMCB, hPage );

            // clear the dirty bit in the page table

            if (dwLinPage = ((PPAGEDATA) hPage) -> PDS_linear_page)
            {
               BYTE   bValue ;
               PBYTE  pByte ;

               vmmModifyPageBits( hVM, dwLinPage, 1, -1, 0, PG_IGNORE, 0 ) ;

               // Touch page to set its accessed bit so that we
               // don't think that we are in a dead lock situation at
               // page fault time.  Basically, in the unmap page routine
               // we check to see if we are unmapping a page that hasn't
               // been accessed.  If we are processing a page fault,
               // then that means that the page is probably being
               // unmapped because it is being demanded to satisfy a
               // request for a different page.  If this is the case,
               // then we assume that we've reached a dead lock situation
               // where we are trying to satisfy requests for 2 video
               // pages at the same time when there is only 1 page of
               // video memory available.  To break out of this deadlock
               // situation we force the VM to full screen.	If we don't
               // preserve the access bit here, then we can wrongly
               // detect that this page hasn't been accessed and convert
               // a VM to fullscreen.

               pByte =
                  (PBYTE) ((dwLinPage << 12) +
                              ((PVMMCB) hVM) -> CB_High_Linear) ;
               bValue = *pByte ;
            }

            pVMCB -> VDD_PageFlags[ i ] |= fPageSaved ;
   
            VDD_VMMem_RecordVRAMMod( pVMCB, i ) ;
         }
      }
      else
      {
         UINT   uPageNum ;
         DWORD  dwPTE ;

         uPageNum = VDD_PHMem_CalcPageAddr( hVM, pVMCB, i ) ;

         if (uPageNum < 0xC0)
         {
            dwPTE = VDD_VMMem_GetPTE( hVM, uPageNum ) ;
            if (dwPTE & P_DIRTY)
            {
               BYTE   bValue ;
               PBYTE  pByte ;

               vmmModifyPageBits( hVM, uPageNum, 1, -1, 0, PG_IGNORE, 0 ) ;

               // touch the page... see above for details.

               pByte =
                  (PBYTE) ((uPageNum << 12) +
                              ((PVMMCB) hVM) -> CB_High_Linear) ;
               bValue = *pByte ;

               // record the saved state...

               pVMCB -> VDD_PageFlags[ i ] |= fPageSaved ;
               VDD_VMMem_RecordVRAMMod( pVMCB, i ) ;
            }
         }
      }
   }

   if (dwVidFlags & fVid_SR_Section)
      VDD_State_EndSRSection() ;

} // VDD_Grab_SaveDirtyPages()

//--------------------------------------------------------------------------
//  
//  DWORD VDD_Grab_GetChgs
//  
//  Description:
//      Force dirty pages to be saved to the VM's save memory,
//      OR dirty bits into the VDD_VRAM_Mods bit array and clear
//      dirty bits.
//  
//  Parameters:
//      DWORD hVM
//         handle to VM
//  
//      PVDDCB pVMCB
//         pointer to VM control block
//
//      PVDDMODSTATE pModState
//         pointer to grabber's mod state buffer
//  
//  Return (DWORD):
//      number of bytes in pModState.VDD_Mod_List
//  
//  History:   Date       Author      Comment
//             10/28/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

DWORD VDD_Grab_GetChgs
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    PVDDMODSTATE    pModState
)
{
   UINT   uVisFirst, uVisPages ;
   DWORD  dwMods ;

   VDD_Grab_SaveDirtyPages( hVM, pVMCB ) ;

   VDD_PHMem_GetVisiblePages( hVM, pVMCB, &uVisFirst, &uVisPages ) ;

   pModState -> VDD_Mod_Flag |= fVDD_M_VRAM ;
   pModState -> VDD_Mod_Count = (WORD) uVisPages ;

   uVisPages += 7 ;
   uVisPages >>= 3 ;

   if (uVisPages > VRAM_Mod_List_Size)
      DEBUGMSG( "Visible pages exceeds VRAM_Mod_List_Size!\r\n" ) ;

   pVMCB -> VDD_Grab_Mods[ 0 ] |= pVMCB -> VDD_VRAM_Mods[ 0 ] ;
   pVMCB -> VDD_VRAM_Mods[ 0 ] = 0 ;

   dwMods = pVMCB -> VDD_Grab_Mods[ 0 ] ;
   dwMods >>= uVisPages ;

   pModState -> VDD_Mod_List[ 0 ] = dwMods ;

   if (dwMods)
   {
      // APP touched it directly...

      pVMCB -> VDD_Mod_Rect = -1 ;
   }
   else
   {
      // for VGA31 grabbers, we keep track of mod_rect.
      // If VDD_Mod_Rect is 0FFFF0000h, then there are no
      // mods. else there are and we return 0 in ecx.

      if (pVMCB -> VDD_Mod_Rect == 0xFFFF0000)
      {
         pModState -> VDD_Mod_Flag &= ~fVDD_M_VRAM ;
         return 0 ;
      }
   }

   return (pModState -> VDD_Mod_Count) ;


} // VDD_Grab_GetChgs()

//---------------------------------------------------------------------------
//  End of File: vddgrab.c
//---------------------------------------------------------------------------

