//---------------------------------------------------------------------------
//
//  Module: vddoem.c
//
//  Purpose:
//     This module is called from various other modules to perform
//     operations specific to an OEM's hardware or the operation is
//     slightly different on different pieces of hardware for one VDD.
//
//     Based on original Windows 3.1 VDD sources as supplied by
//     the Microsoft Windows 3.1 Device Driver Development Kit.
//
//  Development Team:
//     Bryan A. Woodruff
//
//  History:   Date       Author      Comment
//             10/20/93   BryanW      Rewritten in 'C'.
//
//---------------------------------------------------------------------------
//
//               Copyright (c) 1993 Woodruff Software Systems.
//                          All Rights Reserved.
//
//---------------------------------------------------------------------------

#include <vmm.h>
#include "vddvga.h"
#include "portio.h"

#pragma data_seg( "_LDATA", "CODE" ) 

//=============================================================================
//                             L C O D E
//=============================================================================

//--------------------------------------------------------------------------
//  
//  DWORD VDD_OEM_PreRestoreAttr
//  
//  Description:
//      This routine does any programming of the Attribute controller that
//      is not done straight, in line (i.e. starting at register 0 up to
//      register 1Fh).
//  
//  Parameters:
//      DWORD hVM
//         VM handle of VM getting CRTC state (== 0 if VDD)
//  
//      PVDDCB pMemCVMCB
//         MemC owner control block
//  
//      PVDDCB pCRTCVMCB
//         CRTC owner control block
//  
//      WORD wPort
//         3ba or 3da
//  
//      DWORD dwMx
//         mask for output of attribute controller registers
//  
//  Return (DWORD):
//      updated mask for output of attribute controller registers
//  
//  History:   Date       Author      Comment
//             10/27/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

DWORD VDD_OEM_PreRestoreAttr
(
    DWORD           hVM,
    PVDDCB          pMemCVMCB,
    PVDDCB          pCRTCVMCB,
    WORD            wPort,
    DWORD           dwMx
)
{
   UNUSED_PARAM( hVM ) ;
   UNUSED_PARAM( pMemCVMCB ) ;

   // attribute controller regs 10, 12, 13

   VDD_State_AttrOut( FIELDOFFSET( REGSTATE, A_Mode ) -
                      FIELDOFFSET( REGSTATE, A_Pal ),
                      pCRTCVMCB -> VDD_Stt.A_Mode ) ;

   VDD_State_AttrOut( FIELDOFFSET( REGSTATE, A_PlEna ) -
                      FIELDOFFSET( REGSTATE, A_Pal ),
                      pCRTCVMCB -> VDD_Stt.A_PlEna ) ;

   VDD_State_AttrOut( FIELDOFFSET( REGSTATE, A_HPan ) -
                      FIELDOFFSET( REGSTATE, A_Pal ),
                      pCRTCVMCB -> VDD_Stt.A_HPan ) ;

   // attribute controller regs 14, 0-F, 11

   // turn screen off...                    

   wPort = (wPort & 0xFF00) | (pSeqIndx & 0x00FF) ;
   outp( wPort, 1 ) ;
   outp( wPort + 1, (pCRTCVMCB -> VDD_Stt.S_ClMode) | 0x20 ) ;

   // just do first 17 regs

   return (dwMx & 0x1FFFF) ; 

} // VDD_OEM_PreRestoreAttr()

//--------------------------------------------------------------------------
//  
//  VOID VDD_OEM_PostRestoreAttr
//  
//  Description:
//     This routine does any additional Attr or DAC state programming.
//  
//  Parameters:
//      DWORD hVM
//         VM handle of VM getting CRTC state (== 0 if VDD)
//  
//      PVDDCB pMemCVMCB
//         MemC owner control block
//  
//      PVDDCB pCRTCVMCB
//         CRTC owner control block
//  
//      WORD wPort
//         3ba or 3da
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/27/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_OEM_PostRestoreAttr
(
    DWORD           hVM,
    PVDDCB          pMemCVMCB,
    PVDDCB          pCRTCVMCB,
    WORD            wPort
)
{

   UNUSED_PARAM( hVM ) ;
   UNUSED_PARAM( pMemCVMCB ) ;

   VDD_State_AttrOut( FIELDOFFSET( REGSTATE, A_OvScn ) -
                      FIELDOFFSET( REGSTATE, A_Pal ),
                      pCRTCVMCB -> VDD_Stt.A_OvScn ) ;

   // restore clocking mode

   wPort = (wPort & 0xFF00) | (pSeqIndx & 0x00FF) ;
   outp( wPort, 1 ) ;
   outp( wPort + 1, pCRTCVMCB -> VDD_Stt.S_ClMode ) ;

} // VDD_OEM_PostRestoreAttr()

//--------------------------------------------------------------------------
//  
//  BOOL VDD_OEM_AdjustCursor
//  
//  Description:
//      Adjust cursor start and end scan line.
//  
//  Parameters:
//      DWORD hVM
//         handle to VM
//  
//      PVDDCB pVMCB
//         pointer to VM control block
//  
//      PUINT puCurStart
//         cursor start line
//  
//      PUINT puCurEnd
//         cursor end line
//  
//  Return (BOOL):
//      TRUE if cursor visible
//
//      *puCurStart and *puCurEnd are adjusted if necessary.
//  
//  History:   Date       Author      Comment
//             10/27/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

BOOL VDD_OEM_AdjustCursor
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    PUINT           puCurStart,
    PUINT           puCurEnd
)
{
   BYTE  bCurEnd, bCurMax, bCurStart ;

   bCurMax = (BYTE) (*puCurEnd >> 8) ;
   bCurEnd = (BYTE) *puCurEnd ;
   bCurStart = (BYTE) *puCurStart ;

   bCurMax &= 0x0F ;

   if (bCurStart > bCurMax)
      return FALSE ;

   // if start == end, one line cursor

   if (bCurStart == bCurEnd)
      bCurEnd++ ;

   if (bCurEnd == 0)
      bCurEnd = bCurMax + 1 ;

   if (bCurStart > bCurEnd)
   {
      BYTE  bTemp ;

      bTemp = bCurEnd ;
      bCurEnd = bCurStart ;
      bCurStart = bTemp ;
   }
   if (bCurStart > bCurMax)
   {
      bCurStart = 0 ;
      bCurEnd = bCurMax + 1 ;
   }

   *puCurStart = bCurStart ;
   *puCurEnd = ((UINT) bCurMax << 8) | bCurEnd ;

   return TRUE ;

} // VDD_OEM_AdjustCursor()

//---------------------------------------------------------------------------
//  End of File: vddoem.c
//---------------------------------------------------------------------------
