//---------------------------------------------------------------------------
//
//  Module: vmemtrap.c
//
//  Purpose:
//     Allocates memory for each VM and "instances" a page.
//
//  Development Team:
//     Bryan A. Woodruff
//
//  History:   Date       Author      Comment
//              6/10/93   BryanW      Wrote it.
//
//---------------------------------------------------------------------------
//
//               Copyright (c) 1993 Woodruff Software Systems.
//                          All Rights Reserved.
//
//---------------------------------------------------------------------------

#include <vmm.h>
#include "vmemtrap.h"

#pragma data_seg( "_LDATA", "CODE" )

//=============================================================================
//          V I R T U A L   D E V I C E   D E C L A R A T I O N
//=============================================================================

DDB VMEMTRAP_DDB = { NULL,                  // reserved, must be NULL
                     DDK_Version,           // reserved, DDK_Version
                     VMEMTRAP_Device_ID,    // Device ID
                     VMEMTRAP_Major_Ver,    // Major Version
                     VMEMTRAP_Minor_Ver,    // Minor Version
                     NULL,
                     "VMEMTRAP",
                     VDD_Init_Order + 1,
                     (DWORD) vmmwrapVxDControlProc,
                     NULL,
                     NULL,
                     NULL,
                     NULL,
                     NULL,
                     NULL,
                     NULL } ;

//
// This table is used by the vmmwrapVxDControlProc defined in VMMWRAP.ASM.                  
// It lists the messages and associated dispatch functions, it
// must be terminated with -1 and NULL.                  
//

DISPATCHINFO alpVxDDispatchProcs[] =
   { Create_VM,           VMEMTRAP_Create_VM,
     Destroy_VM,          VMEMTRAP_Destroy_VM,
     Device_Init,         VMEMTRAP_Device_Init,
     -1,                  NULL } ;

// Control block offset

UINT guCBOffset ;

// Hooked page reference

WORD gwHookedPage ;

// VMEMTRAP owner

DWORD ghOwnerVM = NULL ;

// Pointer to page-fault handler thunk
   
V86PHPROC pVMEMTRAP_PFault ;

//------------------------------------------------------------------------
//  BOOL VMEMTRAP_Create_VM
//
//  Description:
//     Notification of when VMs (other than system VM) are created.
//
//  Parameters:
//     hVM
//        VM handle
//
//     pCRS
//        pointer to client register structure
//
//  History:   Date       Author      Comment
//              3/ 9/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

BOOL CDECL VMEMTRAP_Create_VM
(
   DWORD    hVM,
   PCRS_32  pCRS
)
{
   PVMCB  pVMCB ;

   UNUSED_PARAM( pCRS ) ;

   vmmTraceOut( "VMEMTRAP_Create_VM\r\n" ) ;

   // set up VM control block pointer

   pVMCB = (PVMCB) (hVM + guCBOffset) ;

   // Allocate a page to map into V86 space

   vmmPageAllocate( &pVMCB -> hPage, 1, PG_HOOKED, hVM,
                    NULL, NULL, NULL, NULL, PageZeroInit ) ;

   // Modify the page bits in VM to specify not present and hooked.

   vmmModifyPageBits( hVM, gwHookedPage, 1, ~P_AVAIL,
                      NULL, PG_HOOKED, NULL ) ;

   return ( TRUE ) ;
   
} // end of VMEMTRAP_Create_VM()

//------------------------------------------------------------------------
//  BOOL VMEMTRAP_Destroy_VM
//
//  Description:
//     Frees memory allocated for "instance" data.
//
//  Parameters:
//     DWORD hVM
//        handle to VM
//
//  History:   Date       Author      Comment
//              6/12/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

BOOL CDECL VMEMTRAP_Destroy_VM
(
   DWORD    hVM
)
{
   PVMCB  pVMCB ;

   vmmTraceOut( "VMEMTRAP_Destroy_VM\r\n" ) ;

   // set up VM control block pointer

   pVMCB = (PVMCB) (hVM + guCBOffset) ;

   vmmPageFree( pVMCB -> hPage, NULL ) ;

   if (hVM == ghOwnerVM)
      ghOwnerVM = NULL ;

   return ( TRUE ) ;

} // end of VMEMTRAP_Destroy_VM()

//------------------------------------------------------------------------
//  VOID VMEMTRAP_PFault
//
//  Description:
//     Resolves the page fault by mapping memory allocated
//     during VM initialization.
//
//  Parameters:
//     DWORD hVM
//        handle to VM causing fault
//
//     UINT uPage
//        linear page causing fault
//
//  History:   Date       Author      Comment
//              6/12/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID CDECL VMEMTRAP_PFault
(
   DWORD    hVM,
   UINT     uPage
)
{
   PVMCB  pVMCB ;

   vmmTraceOutParam( "Page fault for #EAX\r\n", uPage ) ;

   if (gwHookedPage != uPage)
   {
      vmmDebugOut( "Strange! page fault for non-hooked page!\r\n" ) ;
      return ;
   }

   vmmBeginCriticalSection( hVM,
                            Block_Svc_If_Ints_Locked | Block_Enable_Ints ) ;

   if (ghOwnerVM)
   {
      vmmMapIntoV86( vmmGetNulPageHandle(), ghOwnerVM,
                     gwHookedPage, 1, 0, NULL ) ;
      vmmModifyPageBits( ghOwnerVM, gwHookedPage, 1, ~P_AVAIL, NULL,
                         PG_HOOKED, NULL ) ;
   }

   pVMCB = (PVMCB) (hVM + guCBOffset) ;

   vmmMapIntoV86( pVMCB -> hPage, hVM, gwHookedPage, 1, 0, NULL ) ;
   ghOwnerVM = hVM ;

   vmmEndCriticalSection( hVM ) ;

} // end of VMEMTRAP_PFault()

//---------------------------------------------------------------------------
//  End of File: vmemtrap.c
//---------------------------------------------------------------------------
